/**********************************************************

mraynotes.h

Purpose:
The functions here are to be used in writing mental ray
shaders in a way analogous to that described in Steve May's
RManNotes. I found the RManNotes a great resource for
learning shader writing, and I wanted to translate the
approach decsribed there to mental ray.

Author: Brad Christensen with reference to Steve May's RManNotes.

Note:	This header file was made referencing 'rmannotes.sl'
		and the RManNotes which can be found here:

		http://www.cgrg.ohio-state.edu/~smay/RManNotes

The functions defined in this file try to match those useful
in creating the types of shaders described at the website
above. Mainly, these will be mental ray procedural texture
shaders based on regular and stochastic patterns. And of 
course what ever else one can dream up.

The main differences between this file, and rmannotes.sl:

	Generally, function/macro names are kept the same with
	the exception of being having 'mrn_' added to the
	beginning of the function name.

	This file implements the smoothstep function,
	as I could not find any analogous function in the mr
	library. Further more, the smoothstep function uses 
	mental ray's built in mi_spline function for interpolation
	(a cardinal curve instead of what I think is a hermite
	interpolation described in the RManNotes, if there is
	any difference).

	This file implements the function ptlined defining the
	shortest distance from a point to a line segment.

	This file implements lerp and blend, as I could not find 
	any analogous function in the mr library.

	There are noise helper functions that provide signed 
	noise as well as uniformly distributed noise in a 
	specified range.

\***********************************************************/

#ifndef __MRAYNOTES_H__
#define __MRAYNOTES_H__

#include <math.h>
#include <shader.h>

 /*********************************************************\
Transition/Interpolation Functions
 \*********************************************************/

static inline miScalar
mrn_smoothstep(miScalar a, miScalar b, miScalar x)
{
	if(x < a)
		return 0;
	else if(x > b)
		return 1;
	else
	{
		// return the cardinal interpolation between a and b
		// using the built in mi_spline(x,n,ctl)

		// we want the curve to go from 0 to 1.
		miScalar pts[2] = {0.0,1.0};
		// Call the function such that 0 < x < 1
		return mi_spline((x-a)/(b-a),2,pts);
	}
}

static inline miScalar
mrn_pulse(miScalar a, miScalar b, miScalar fuzz, miScalar x)
{
	return (mrn_smoothstep((a-fuzz),a,x) - mrn_smoothstep((b-fuzz),b,x));
}

 /*********************************************************\
Tiling/Repeat Funcions
 \*********************************************************/

static inline miScalar
mrn_repeat(miScalar x, miScalar freq)
{
	return (miScalar)(fmod( (x*freq), 1.0 ));
}

static inline miScalar
mrn_whichtile(miScalar x, miScalar freq)
{
	return (miScalar)(floor(x*freq));
}

static inline miBoolean
mrn_odd(miScalar x)
{
	return (miBoolean)( fmod(x,2.0) == 1 );
}

static inline miBoolean
mrn_even(miScalar x)
{
	return (miBoolean)( fmod(x,2.0) == 0 );
}

 /*********************************************************\
Boolean Operations (Perlin85)
Each returns a value between 0 and 1. All inputs should also
be between 0 and 1
 \*********************************************************/

static inline miScalar
mrn_intersection(miScalar a, miScalar b)
{

	return ( a*b );

}

static inline miScalar
mrn_union(miScalar a, miScalar b)
{
	return ( (a+b) - (a*b) );
}

static inline miScalar
mrn_difference(miScalar a, miScalar b)
{
	return ( (a) - (a*b) );
}

static inline miScalar
mrn_complement(miScalar a)
{
	return (1-a);
}

 /*********************************************************\
Noise Helper Functions
 \*********************************************************/

// 1d signed noise in the range -1,1
static inline miScalar
mrn_snoise_1d(miScalar x)
{
	return (mi_noise_1d(x) * 2 - 1);
}

// 2d signed noise in the range -1,1
static inline miScalar
mrn_snoise_2d(miScalar x, miScalar y)
{
	return (mi_noise_2d(x,y) * 2 - 1);
}

// 3d signed noise in the range -1,1
static inline miScalar
mrn_snoise_3d(miVector *x)
{
	return (mi_noise_3d(x) * 2 - 1);
}

// 1d uniformly distributed noise within the range lo to hi
static inline miScalar
mrn_udn_1d(miScalar x, miScalar lo, miScalar hi)
{
	return mrn_smoothstep(0.25f, 0.75f, mi_noise_1d(x)) * ( (hi - lo) + lo );
}

// 2d uniformly distributed noise within the range lo to hi
static inline miScalar
mrn_udn_2d(miScalar x, miScalar y, miScalar lo, miScalar hi)
{
	return mrn_smoothstep(0.25f, 0.75f, mi_noise_2d(x,y)) * ( (hi - lo) + lo );
}

// 3d uniformly distributed noise within the range lo to hi
static inline miScalar
mrn_udn_3d(miVector *x, miScalar lo, miScalar hi)
{
	return mrn_smoothstep(0.25f, 0.75f, mi_noise_3d(x)) * ( (hi - lo) + lo );
}

 /*********************************************************\
Miscellaneous Stuff
 \*********************************************************/

// Distance between two points
static inline miScalar
mrn_distance(miVector *a, miVector *b)
{
	miVector r;
	mi_vector_sub(&r,a,b);
	return (miScalar)sqrt( mi_vector_dot(&r,&r) );
}

// A function to calculate the shortest distance from a
// point to a line. s and d are the points of the line, and
// p is the point we want to calculate the distance to the 
// line.
static inline miScalar
mrn_ptlined(miVector *s, miVector *d, miVector *p)
{
	// working variables. 
	// sd = vector from s to d
	// sdn = normalized vector from s to d
	// sp = vector from s to p
	// spp = vector from s to projected point on sd
	// r = vector from projected point to p
	miVector sd, sdn, sp, spp, r;// = {0.0,0.0,0.0};
	miScalar lengthsd, lengthspp;

	mi_vector_sub(&sd,d,s); // sd is a vector from s to d
	mi_vector_sub(&sp,p,s); // sp is a vector from s to p

	// get normalized version of sd as well as sd's length
	sdn = sd;
	mi_vector_normalize(&sdn);
	lengthsd = mi_vector_norm(&sd);

	// calculate the vector from s to the projection of p 
	// onto sd. Use spp to hold the new vector
	lengthspp = mi_vector_dot(&sp,&sdn);
	mi_vector_mul(&sdn,lengthspp); // extend sdn to the projected point
	spp = sdn;

	// Clamp spp in the range of s to d
	// Because sdn is normalized (length is 1) along the line, 
	// lengthspp must be the length of the projected vector along 
	// the same line. Therefor, if lengthspp is greater than the 
	// length of sd, clamp spp to sd, otherwise if lengthspp is
	// lessthan 0, clamp to s of sd, which is 0,0,0.
	if(lengthspp < 0.0f) // clamp the vector spp to s
	{
		spp.x = 0.0f;
		spp.y = 0.0f;
		spp.z = 0.0f;
	}
	else if(lengthspp > lengthsd)
	{
		spp = sd;
	}

    // calculate the distance from sp to spp
	mi_vector_sub(&r,&sp,&spp);
	return mi_vector_norm(&r); // return the length of the result vector r
}

// 2D Rotation of point (x,y) about origin (ox,oy) by an
// angle rad. The resulting point is (rx, ry).
static inline void
mrn_rotate2d(miScalar x, miScalar y, miScalar rad, miScalar ox, miScalar oy, miScalar *rx, miScalar *ry)
{
	*rx = ((x) - (ox)) * (miScalar)cos(rad) - ((y) - (oy)) * (miScalar)sin(rad) + (ox);
	*ry = ((x) - (ox)) * (miScalar)sin(rad) + ((y) - (oy)) * (miScalar)cos(rad) + (oy);
}

// Convert degrees to radians
static inline miScalar
mrn_radians(miScalar degrees)
{
	return (miScalar)( degrees * (M_PI / 180) );
}

// 2D cartesian -> polar coordinates
// converts the point (x,y) to radius 'r' and angle 'theta'
// (in radians). theta will be in the range [-PI,PI].
static inline void
mrn_topolar2d(miScalar x, miScalar y, miScalar *r, miScalar *theta)
{
	*r = (miScalar)sqrt((x) * (x) + (y) * (y));
	*theta = (miScalar)atan2f(y,x);
}

// clamp a scalar to range. lo <= s <= hi
static inline miScalar
mrn_clamp(miScalar s, miScalar lo, miScalar hi)
{
	if((lo <= s) && (s <= hi))
		return s;
	else if(s < lo)
		return lo;
	else
		return hi;
}

// Linear Interpolation
static inline miScalar
mrn_lerp(miScalar a, miScalar b, miScalar x)
{
	return ( (a*(1 - x)) + (b*x) );
}

// Blend colors b over a component wise using lerp
static inline miColor
mrn_blend(miColor a, miColor b, miScalar x)
{
	miColor r;
	r.r = mrn_lerp(a.r,b.r,x);
	r.g = mrn_lerp(a.g,b.g,x);
	r.b = mrn_lerp(a.b,b.b,x);
	r.a = mrn_clamp(a.a + b.a,0.0f,1.0f);
	return r;
}

#endif // __MRAYNOTES_H__