/******************************************************************************\
| SHADERHLP.H																   |
|																			   |
| Little helper functions for writing mental ray shaders.					   |
|																			   |
| Author:	Daniel "thebigMuh" Rind											   |
| History:	28.03.2003	Initial writing. A few functions for modifying colors, |
|						and a function for mapping 2 points in [0, 1) to a	   |
|						sphere (usefull for generating unit direction vectors  |
|						from MR's sample function)							   |
|			08.04.2003  Added the simple mihRAY_SECILLUM macro for checking	   |
|						state->type for secondary illumination rays.		   |
|			19.04.2003	Added mih_find_refraction_index function for finding   |
|						the correct IOR by walking the state hierarchy.	This   |
|						function was taken from the mental images base shader  |
|						library.											   |
|						Also added clamping macros, and added functions that   |
|						take values instead of references.					   |
|			21.04.2003	Added mih_rgb_transfer to generate a color based	   |
|						on a simple scalar input, so that the avg of n evenly  |
|						distributed scalars from [0...1] will always be the	   |
|						same. This is used in my diffraction shader.		   |
|																			   |
\******************************************************************************/

#ifndef __SHADERHLP_H__
#define __SHADERHLP_H__

#include <math.h>
#include "shader.h"

/******************************************************************************\
| COLOR FUNCTIONS															   |
\******************************************************************************/

// sets a color to 4 scalar values
static inline void 
mih_color_set(miColor *n, 
			  miScalar r = 0, 
			  miScalar g = 0, 
			  miScalar b = 0, 
			  miScalar a = 0)
{
	n->r = r;
	n->g = g;
	n->b = b;
	n->a = a;
}

// add two colors
static inline void 
mih_color_add(miColor *r, 
			  miColor *a, 
			  miColor *b)
{
	r->r = a->r + b->r;
	r->g = a->g + b->g;
	r->b = a->b + b->b;
	r->a = a->a + b->a;
}

// add a scalar to a color
static inline void 
mih_color_add(miColor *r, 
			  miColor *a, 
			  miScalar *b)
{
	r->r = a->r + *b;
	r->g = a->g + *b;
	r->b = a->b + *b;
	r->a = a->a + *b;
}

// add a scalar to a color (byval)
static inline void 
mih_color_add(miColor *r, 
			  miColor *a, 
			  miScalar b)
{
	r->r = a->r + b;
	r->g = a->g + b;
	r->b = a->b + b;
	r->a = a->a + b;
}

// multiply two colors
static inline void 
mih_color_mul(miColor *r, 
			  miColor *a, 
			  miColor *b)
{
	r->r = a->r * b->r;
	r->g = a->g * b->g;
	r->b = a->b * b->b;
	r->a = a->a * b->a;
}

// multiply a scalar and a color
static inline void 
mih_color_mul(miColor *r, 
			  miColor *a, 
			  miScalar *b)
{
	r->r = a->r * *b;
	r->g = a->g * *b;
	r->b = a->b * *b;
	r->a = a->a * *b;
}

// multiply a scalar and a color (byval)
static inline void 
mih_color_mul(miColor *r, 
			  miColor *a, 
			  miScalar b)
{
	r->r = a->r * b;
	r->g = a->g * b;
	r->b = a->b * b;
	r->a = a->a * b;
}

// divide two colors
static inline void 
mih_color_div(miColor *r, 
			  miColor *a, 
			  miColor *b)
{
	r->r = a->r / b->r;
	r->g = a->g / b->g;
	r->b = a->b / b->b;
	r->a = a->a / b->a;
}

// divide a color through a scalar
static inline void 
mih_color_div(miColor *r, 
			  miColor *a, 
			  miScalar *b)
{
	r->r = a->r / *b;
	r->g = a->g / *b;
	r->b = a->b / *b;
	r->a = a->a / *b;
}

// divide a color through a scalar (byval)
static inline void 
mih_color_div(miColor *r, 
			  miColor *a, 
			  miScalar b)
{
	r->r = a->r / b;
	r->g = a->g / b;
	r->b = a->b / b;
	r->a = a->a / b;
}

// mix two colors a and b based on a scalar c, where 0 means full color a, and
// 1 means full color b gets applied
static inline void 
mih_color_mix(miColor *r, 
			  miColor *a, 
			  miColor *b, 
			  miScalar c)
{
	r->r = a->r * (1.0f - c) + b->r * c;
	r->g = a->g * (1.0f - c) + b->g * c;
	r->b = a->b * (1.0f - c) + b->b * c;
	r->a = a->a * (1.0f - c) + b->a * c;
}

/******************************************************************************\
| MACROS																	   |
\******************************************************************************/

// simple macro for use in if() statements. checks if _raytype is a secondary
// illumination ray - either a photon, or a final gather ray.
#define mihRAY_SECILLUM(_raytype)					\
	((_raytype) == miRAY_FINALGATHER ||				\
	 miRAY_PHOTON(_raytype) == miTRUE)

// clamp a scalar from 0 to 1
#define mih_clamp_0_1(_value)						\
	((_value) > 1.0f) ? 1.0f : ( ((_value) < 0.0f) ? 0.0f : _value )

/******************************************************************************\
| MISCELLANEOUS FUNCTIONS													   |
\******************************************************************************/

// map 2 values in the range [0, 1) to the hull of a sphere
// this is the slowest function in my shaders, so if anyone has any ideas on
// how to improve this, please mail me!
static inline miVector 
mih_sphere_sample(miGeoScalar sample[2])
{
	// Compute a unit direction vector from 2 samples on a spherical hull:
	miVector	smp;
	double		t, st, phirad;

	t = 2.0 * sample[0] - 1.0;			// map from [0, 1) to [-1, 1)
	
	st = sqrt(1.0 - t*t);

	phirad = sample[1] * 2.0 * M_PI;	// map from [0, 1) to [0, 2pi)

	smp.x = (miScalar)(st * cos(phirad));
	smp.y = (miScalar)(st * sin(phirad));
	smp.z = (miScalar)t;

	return smp;
}

// find the actual IOR to apply by walking the state hierarchy
static inline miBoolean 
mih_find_refraction_index(
	miState		*state,
	miScalar	*ior_in,
	miScalar	*ior_out,
	miScalar	mtl_ior)
{
	register int		num = 0;
	register miState	*s, *s_in = NULL;	/* for finding enclosing mtl */

	for (s = state; s; s = s->parent)			/* history? */
	{
		if ((s->type == miRAY_TRANSPARENT		||
		     s->type == miRAY_REFRACT			||
		     s->type == miPHOTON_TRANSMIT_SPECULAR	||
		     s->type == miPHOTON_TRANSMIT_GLOSSY	||
		     s->type == miPHOTON_TRANSMIT_DIFFUSE) &&
		     s->parent && s->parent->shader == state->shader) 
		{
			num++;
			if (!s_in)
				s_in = s->parent;
		}
	}

	if (!(num & 1)) 
	{				/* even: entering */
		*ior_out = mtl_ior;
		*ior_in	= state->parent && state->parent->ior != 0 ?
							state->parent->ior : 1;
		if (!state->refraction_volume)
			state->refraction_volume = state->volume;
	} 
	else 
	{					/* odd: exiting */
		*ior_in	= mtl_ior;
		*ior_out = s_in && s_in->ior_in != 0 ? s_in->ior_in : 1;
		state->refraction_volume = s_in ? s_in->volume : state->camera->volume;
	}

	state->ior_in = *ior_in;
	state->ior    = *ior_out;
	return (miBoolean)(!(num & 1));
}

// generator function for color bands
// behaviour for t not in [0...1] is undefined (but not difficult to guess...)
static inline void
mih_rgb_transfer(miColor *r, miScalar t)
{
	if (t < 0.25)
	{
		r->r = t * 16.0f;
		r->g = 0.0f;
		r->b = 0.0f;
		return;
	}
	if (t < 0.5)
	{
		r->r = t * -16.0f + 8.0f;
		r->g = t * 16.0f - 4.0f;
		r->b = 0.0f;
		return;
	}
	if (t < 0.75)
	{
		r->r = 0.0f;
		r->g = t * -16.0f + 12.0f;
		r->b = t * 16.0f - 8.0f;
		return;
	}
	r->r = 0.0f;
	r->g = 0.0f;
	r->b = t * -16.0f + 16.0f;
}

#endif //__SHADERHLP_H__